/*! \file playwave.c \brief Wave Player */
//###########################################################
///	\code #include "dos.h" \endcode
///	\par Uebersicht
//###################################################################################
//
// Play wave files from SD-Card
//
// Uses 8 bit R2R-DAC on DAC_PORT, or PWM at PD7 (or both;)
//
// Does not play stereo files, files with 16 bit samples or files with compression.
//
// Tested from 8000Hz to 44100Hz samplerates at 8MHz and 16MHz cpu clock.
//
// SD card connection on MISO, MOSI, SCK.
// CS of the SD card is connected to SS pin.
//
//#########################################################################
// Last change: 26.11.2010
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: AVR-GCC 4.1.1
//#########################################################################
//@{
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

#include "..\dos.h"
#include "playwave.h"

volatile uint8_t play_buffer_index, play_buffer_number;
uint8_t play_buffer0[PLAY_BUFFER_SIZE], play_buffer1[PLAY_BUFFER_SIZE];

//###################################################################################
// Audio output ISR
ISR(TIMER1_COMPA_vect)
//###################################################################################
{

 if(play_buffer_number & 1)
 {
#if R2R_DAC
	DAC_PORT = play_buffer1[play_buffer_index];
#endif
#if PWM_DAC
	OCR2 = play_buffer1[play_buffer_index];
#endif
 }
 else
 {
#if R2R_DAC
	DAC_PORT = play_buffer0[play_buffer_index];
#endif
#if PWM_DAC
	OCR2 = play_buffer0[play_buffer_index];
#endif
 }

 play_buffer_index++;

#if PLAY_BUFFER_SIZE == 256
 if(play_buffer_index == 0) play_buffer_number++;
#else
 if(play_buffer_index == PLAY_BUFFER_SIZE)
 {
  play_buffer_index = 0;
  play_buffer_number++;
 } 
#endif
}

//###################################################################################
void PlayWave(char *name)
//###################################################################################
{
 unsigned char result;
 unsigned int read_bytes;
 
 result=Fopen(name,'r');
 if(result==F_OK)
  {
	uint16_t sample_rate;

	sample_rate = ReadWaveHeader();
    if(sample_rate == 0) { Fclose(); return; }

    OCR1A = ((F_CPU / 8) / sample_rate) - 1; // set the samplerate for timer1

	play_buffer_index = 0; 
	play_buffer_number = 0;

    // fill first audio buffer
	read_bytes = Fread(play_buffer0,PLAY_BUFFER_SIZE);

	StartAudioTimer(); // start playing

	do
	{
	 if(play_buffer_number & 1) // buffer number 1 is playing, read buffer 0
	 {
	   read_bytes = Fread(play_buffer0,PLAY_BUFFER_SIZE);
       while(play_buffer_number & 1); // ugly waiting loop, try to use a state machine
	 }
	 else  // buffer number 0 is playing, read buffer 1
	 {
	   read_bytes = Fread(play_buffer1,PLAY_BUFFER_SIZE);
       while((play_buffer_number & 1) == 0);
	 }

	} while(read_bytes == PLAY_BUFFER_SIZE);
	// if read_bytes is smaller PLAY_BUFFER_SIZE ignore these samples
	// or fill buffer with zero samples (0x80) at the end
	
	StopAudioTimer();
	Fclose();
  }
}

//###################################################################################
// Reads the wave header
// Returns a samplerate if successfull, zero if there is an error.
uint16_t ReadWaveHeader(void)
//###################################################################################
{
	uint32_t chunksize;
	uint16_t read_bytes, sample_rate = 0;
	uint8_t by;

	do
	{
		read_bytes = Fread(play_buffer0, 4); // read a chunk header
		
		if(play_buffer0[0] == 'R' && play_buffer0[1] == 'I' && play_buffer0[2] == 'F' && play_buffer0[3] == 'F')
		{
			read_bytes = Fread(play_buffer0, 4); // skip the chunk size
			continue;
		}

		if(play_buffer0[0] == 'W' && play_buffer0[1] == 'A' && play_buffer0[2] == 'V' && play_buffer0[3] == 'E')
		{
			continue;
		}


		if(play_buffer0[0] == 'f' && play_buffer0[1] == 'm' && play_buffer0[2] == 't' && play_buffer0[3] == ' ')
		{
			read_bytes = Fread(play_buffer0, 4); // read the chunk size
			chunksize = play_buffer0[0] + ((uint32_t)play_buffer0[1] << 8) + ((uint32_t)play_buffer0[2] << 16) + ((uint32_t)play_buffer0[3] << 24);
			
			read_bytes = Fread(play_buffer0, 2); // read format PCM, ADPCM...
			if(play_buffer0[0] != 1 || play_buffer0[1] != 0) return 0;

			read_bytes = Fread(play_buffer0, 2); // read modus MONO, STEREO...
			if(play_buffer0[0] != 1 || play_buffer0[1] != 0) return 0;

			read_bytes = Fread(play_buffer0, 4); // read the samplerate
			sample_rate = play_buffer0[0] + ((uint16_t)play_buffer0[1] << 8);

			read_bytes = Fread(play_buffer0, 4); // byte_per_sec

			read_bytes = Fread(play_buffer0, 2); // byte_per_sample
			if(play_buffer0[0] != 1 || play_buffer0[1] != 0) return 0;

			read_bytes = Fread(play_buffer0, 2); // bits_per_sample
			if(play_buffer0[0] != 8 || play_buffer0[1] != 0) return 0;
			chunksize -= 16; 

			// skip unknown parts of the fmt chunk
			while(chunksize-- && read_bytes != 0) read_bytes = Fread(&by, 1);

			continue;
		}

		if(play_buffer0[0] == 'd' && play_buffer0[1] == 'a' && play_buffer0[2] == 't' && play_buffer0[3] == 'a')
		{
			read_bytes = Fread(play_buffer0, 4); // read the chunk size
			chunksize = play_buffer0[0] + ((uint32_t)play_buffer0[1] << 8) + ((uint32_t)play_buffer0[2] << 16) + ((uint32_t)play_buffer0[3] << 24);

			return sample_rate;
		}

		// if we come here, we found an unknown chunk (like "fact")
		// read the chunksize and skip the data
		read_bytes = Fread(play_buffer0, 4);
		chunksize = play_buffer0[0] + ((uint32_t)play_buffer0[1] << 8) + ((uint32_t)play_buffer0[2] << 16) + ((uint32_t)play_buffer0[3] << 24);
		while(chunksize-- && read_bytes != 0) read_bytes = Fread(&by, 1);

	} while(read_bytes != 0);

	return 0;
}

//###################################################################################
void StartAudioTimer(void)
//###################################################################################
{
 // CTC Mode, Clkio/8 -> 2MHz at 16MHz F_CPU 
 TCCR1A = (0<<COM1A1) | (0<<COM1A0) | (0<<WGM11) | (0<<WGM10);
 TCCR1B = (0<<WGM13) | (1<<WGM12) | (0<<CS12) | (1<<CS11) | (0<<CS10);

#if R2R_DAC
 DAC_DDR = 0xFF; 	// All outputs
 DAC_PORT = 0x80;   // Zero line for 8bit audio. If it does a click here, use a loop for counting up DAC_PORT.
#endif

#if PWM_DAC
 // Fast PWM, no prescaler 
 TCCR2 = (1<<COM21) | (1<<COM20) | (1<<WGM21) | (1<<WGM20) | (1<<CS20);
 DDRD |= (1<<PD7);
#endif

#if defined (__AVR_ATmega644__) || defined (__AVR_ATmega168__)
 TIMSK1 |= (1<<OCIE1A); 
 TIFR1  |= (1<<OCF1A); 
#endif

#if defined (__AVR_ATmega32__)
 TIMSK |= (1<<OCIE1A);
 TIFR  |= (1<<OCF1A); 
#endif
}

//###################################################################################
void StopAudioTimer(void)
//###################################################################################
{
#if defined (__AVR_ATmega644__) || defined (__AVR_ATmega168__)
 TIMSK1 &= ~(1<<OCIE1A); 
#endif

#if defined (__AVR_ATmega32__)
 TIMSK &= ~(1<<OCIE1A);
#endif
}

//@}
